// Open Source, private use only! 
// "Lokadressen melden" nach DiMo (Digitale Modellbahn)
// Heft 3-2020
// Autor F. Bollow

// Quellcode für Arduino Nano und Arduino UNO

// Initiale Zuordnung Display/Abschnitt ist 1, 2, 3, 4, 5, 6
// Beim Start geben die Displays ihre Abschnittsnummer aus, ca 2sek.
// Mit: nr  5, nr 8, nr 12, nr 23 etc.
// Solange keine Meldung empfangen wurde, ist die Anzeige: -
// dann clr/Occu/"Nummer"


#define SWstand  26

#include <EEPROM.h>  
// EEPROM Belegung:

#define EEPMelder 3
#define EEPMarray 4   // Belegung 7 Einträge, Zuordung Dspl/Melder
#define MarrayDIM 7

#define DISPLAYS  6

#include <TM1637Display.h>  
#define CLK_2        2
#define DIO_3        3
#define CLK_4        4
#define DIO_5        5
#define CLK_6        6
#define DIO_7        7
#define CLK_8        8 
#define DIO_9        9
#define CLK_10      10
#define DIO_11      11
#define CLK_12      12
#define DIO_13      13

#define TasteTm     A5
#define TasteTs     A4
  
TM1637Display display1(CLK_2, DIO_3); // instantiate an object
TM1637Display display2(CLK_4, DIO_5); // instantiate an object
TM1637Display display3(CLK_6, DIO_7); // instantiate an object
TM1637Display display4(CLK_8, DIO_9); // instantiate an object
TM1637Display display5(CLK_10, DIO_11); // instantiate an object
TM1637Display display6(CLK_12, DIO_13); // instantiate an object

// Display, Pin-Reihenfolge
// GND sw
// VCC rt
// DIO ws
// CLK ge

// Display-Zuordnung auf Belegtmeldeabschnitte, aus EEPROM lesen
// Index ist Display 1 bis 6
// Das Array wird beim Erkennen von Abschnittsmeldungen gescannt, ob
// eine Übereinstimmung vorhanden ist.
// Index (1..6) ist Display Nr., Inhalt der Abschnitt

uint8_t BelegtNum[MarrayDIM] = {0,1,2,3,4,5,6}; // Werte sind 1...24
// Array nur bei EEPROM leer bzw. bei Tasteneingabe Tm/Ts wirksam!
// Index 0 ist Display Nr. für Tastenwahl

// write/read Queue
#define QDIM 64
#define QPMSK QDIM-1
volatile char Queue[QDIM] = {0};
volatile uint16_t WRptr  =  0;
volatile uint16_t RDptr  =  0; 

volatile uint8_t x;
volatile uint8_t d; 

//    01
//  20  02
//    40
//  10  04
//    08  80 dp: displayL.setSegments(data)


uint8_t Bmelder    = 1;
uint8_t MelderNr   = 1;
uint8_t ChangeState = 0;
uint8_t DsplNrHold = 1;

enum MsgNr {EMPTY, OCCU, CLEAR, NR, OK, DIS};


// -----------------------------------------------------

void setup() {
//Serial.begin(9600);
//Serial.println("Die wahren Helden");

 pinMode(TasteTm,INPUT_PULLUP); 
 pinMode(TasteTs,INPUT_PULLUP); 

display1.setBrightness(2);
display2.setBrightness(2);
display3.setBrightness(2);
display4.setBrightness(2);
display5.setBrightness(2);
display6.setBrightness(2);

for (int k=1; k<=DISPLAYS; k++) DisplayMsg(k, EMPTY);
DisplayNumber(1,SWstand,true);
delay(500);

 //pinMode(13, OUTPUT);
 //digitalWrite(13, LOW);
  
  //--configure UART0 of nano: Bd = 19200; RXRDY interrupt----
  UCSR0B = 0b10011000;  //
  UCSR0C = 0b00000110;  //
  //-Bd: 19200-------------
  UBRR0L = 0x33;
  UBRR0H = 0x00;
  //--global int enable--
  bitSet(SREG, 7);
  //UDR3 = 0x3A;  // Prompt (nur 1 Zeichen)  
  // UDR3 (direkt) nicht zuverlaessig

  //digitalWrite(13, HIGH); //debugging code

//EEPROM.write(EEPMelder,0xEE);  // leer setzen
MelderNr = EEPROM.read(EEPMelder); 

if ((MelderNr > 24) || (MelderNr < 1)) { // EEPROM leer?
    EEPROM.write(EEPMelder,1);
    for (int k = 0; k < MarrayDIM; k++) {
        int EEPadr = EEPMarray + k;
        int EEPval = BelegtNum[k];
        EEPROM.write(EEPadr,EEPval); 
    }
}
else {  // EEPROM akzeptiert, Meldeadressen umladen in RAM
    for (int k = 0; k < MarrayDIM; k++) {
       BelegtNum[k] = EEPROM.read(EEPMarray+k); 
    }
}

DisplayBmNumbers();

delay(3000);
for (int k=1; k<=DISPLAYS; k++) DisplayMsg(k, OK);
delay(500);

uart0dOut(0x20); // alle Melder anfordern

} // setup()


void DisplayBmNumbers(void) {
for (int k=1; k<=DISPLAYS; k++) {
    DisplayMsg(k, NR);
    DisplayNumber(k, BelegtNum[k], 0);
    }  
} // DisplayBmNumbers()


void DisplayNumber(uint8_t Display, uint16_t Numb, uint8_t Digits) {
switch (Display) {
case 1: if (Digits) display1.showNumberDec(Numb, false);
        else        display1.showNumberDec(Numb, false, 2, 2);
       break;
case 2: if (Digits) display2.showNumberDec(Numb, false);
        else        display2.showNumberDec(Numb, false, 2, 2);
       break;
case 3: if (Digits) display3.showNumberDec(Numb, false);
        else        display3.showNumberDec(Numb, false, 2, 2);
       break;
case 4: if (Digits) display4.showNumberDec(Numb, false);
        else        display4.showNumberDec(Numb, false, 2, 2);
       break;
case 5: if (Digits) display5.showNumberDec(Numb, false);
        else        display5.showNumberDec(Numb, false, 2, 2);
       break;
case 6: if (Digits) display6.showNumberDec(Numb, false);
        else        display6.showNumberDec(Numb, false, 2, 2);
       break;
default:
       break;       
  
}
} // DisplayNumber()


void DisplayMsg(uint8_t Display, uint8_t Msg) {

const uint8_t dataempty[] = { 0, 0, 0, 0};   
const uint8_t dataOccu[] = {0x3F,0x58,0x58,0x1C};
const uint8_t dataClr [] = {0x58,0x18,0x50,0x00};
const uint8_t dataNr  [] = {0x54,0x50,0x00,0x00};
const uint8_t dataOk  [] = { 0x40, 0, 0, 0};
const uint8_t data_d  [] = { 0x5E, 0, 0, 0};

const uint8_t *parray [] = {dataempty, dataOccu, dataClr, dataNr, dataOk, data_d};

switch (Display) {
case 1: display1.setSegments(parray[Msg]);  
       break;
case 2: display2.setSegments(parray[Msg]);  
       break;
case 3: display3.setSegments(parray[Msg]);  
       break;
case 4: display4.setSegments(parray[Msg]);  
       break;
case 5: display5.setSegments(parray[Msg]);  
       break;
case 6: display6.setSegments(parray[Msg]);  
       break;
default:
       break;       
}
} // DisplayMsg()


void loop() {
  
  if (analogRead(TasteTm) < 100) {    
     switch (ChangeState) {
     case 0:
            if (analogRead(TasteTs) < 100) {
                ChangeState = 1; // d, DisplayNr Wahl 1...6
                DisplayMsg(1, DIS);
                DisplayNumber(1, DsplNrHold, 0);
                delay(200);
                while (analogRead(TasteTs) < 100) {}; // warten, bis Taste losgelassen
                }
            else {ChangeState = 2; // nr, Belegtmelder Nummer 1...24
                 DisplayMsg(DsplNrHold, NR);
                 DisplayNumber(DsplNrHold, Bmelder, 0);
                 }
            delay(200);
            while (analogRead(TasteTm) < 100) {}; // warten, bis Taste losgelassen
            break;
     case 1: 
            // Display Nr ist der Wert von DsplNrHold  
            DisplayMsg(1, OK);          
            ChangeState = 0;
            delay(200);
            uart0dOut(0x20);  // alle Melder neu anfordern   
            break;
     case 2: 
            // EEP write, Bmelder an Pos. DsplNrHold 
            if (EEPROM.read(EEPMarray+DsplNrHold) != Bmelder) {
                EEPROM.write(EEPMarray+DsplNrHold, Bmelder);
                BelegtNum[DsplNrHold] = Bmelder;
            }
            DisplayMsg(DsplNrHold, OK);
            ChangeState = 0; 
            delay(200);
            uart0dOut(0x20);  // alle Melder neu anfordern      
            break;
     default:
            ChangeState = 0;
            break;          
     } //switch     
  } // if Tm
  else { // Tm nicht gedrueckt
       if (ChangeState) {
          if (analogRead(TasteTs) < 100) { // Taste zum zaehlen
          switch (ChangeState) {
          case 1: 
                 DsplNrHold++;
                 if (DsplNrHold > DISPLAYS) DsplNrHold = 1;
                 display1.showNumberDec(DsplNrHold, false, 2, 2);
                 delay(200);
                 break;
          case 2:
                 Bmelder++;
                 if (Bmelder > 24) Bmelder = 1;
                 DisplayNumber(DsplNrHold, Bmelder, 0);
                 delay(200);
                 break;
          default:
                 ChangeState = 0;
                 break;       
            } // switch
          } // if Ts       
        }
        else fetch(); // keine States, keine Tasten
  }
 } // loop()



uint8_t SearchBMeld(uint8_t BMeld) {  // return value = display

for (int k=1; k <= DISPLAYS;k++) {
    if (BelegtNum[k] == BMeld) return k;
    }
return 0;
} // SearchBMeld()


void DsplFrame(uint8_t dat) {  // dat ist ein Byte aus Queue

static uint8_t  State = 0;
static uint8_t  Hadr = 0;
uint16_t        Ladr = 0;
static uint8_t  ActiveDisp = 0; // das static fehlte

//DisplayNumber(4,RDptr,1); // Debug Info: ist ok

switch (dat) {
case 0xFC: State = 1;   // Frame start
     //digitalWrite(13, HIGH); //debugging code
     break;
case 0xFF: State = 0;   // Frame end
     //digitalWrite(13, LOW); //debugging code    
     break;
default:    
     switch (State) {
     case 1:            // Nummer Belegtmelder 1...24             
             ActiveDisp = SearchBMeld(dat);
             if (ActiveDisp) {                
                State = 2;
             }
             else State = 0;  // Melder not in list           
             break;
     case 2:           // Lokadressteil high
             Hadr = dat & 0x7F;
             State = 3;
             break;
     case 3:           // Lokadressteil low
             Ladr = (Hadr << 8) + dat;   
                 
                switch (Ladr) {
                case 0x7777: 
                          DisplayMsg(ActiveDisp, OCCU);   // displ: ("Occ.");
                          State = 0;
                          break;
                case 0:   DisplayMsg(ActiveDisp, CLEAR);  // displ: ("clr.");
                          State = 0;
                          break;                          
                default:  DisplayNumber(ActiveDisp, Ladr, 1);
                          State = 0;
                          break;    
                }              
             //digitalWrite(13, LOW); //debugging code            
     default:State = 0;
             break;           
     }    
 }
} // DsplFrame()
 

void fetch (void) {

while (WRptr != RDptr)
  {
    d = Queue[RDptr];
    RDptr++;
    RDptr = RDptr & QPMSK;

    DsplFrame(d);
  } // while
} // fetch() 


ISR(USART_RX_vect)  // USART0 does'nt work!
{
  //digitalWrite(13, HIGH); //debugging code
  x = UDR0;    //receive from UART3
  //UDR3 = 0x3E;
  Queue[WRptr] = x;
  WRptr++;
  WRptr = WRptr & QPMSK;
  //if (x==0x23) flag = true;
}

void uart0dOut(uint8_t dat) {
// transmitter empty? (data sheet p.191)
while (!(UCSR0A & 0b00100000)) {}
      UDR0 = dat;
}
 
